#!/usr/bin/env python3

"""Actions for commonly-used elasticsearch operations."""

import os
import sys
from functools import partial
from urllib.parse import urljoin


def _add_path(path):
    if path not in sys.path:
        sys.path.insert(1, path)


_path = os.path.dirname(os.path.realpath(__file__))
_hooks = os.path.abspath(os.path.join(_path, "../hooks"))


_add_path(_hooks)


from charmhelpers.core.hookenv import (  # noqa E402
    function_fail,
    function_get,
    function_set,
    log,
)
from charmhelpers.fetch import install  # noqa E402


try:
    import requests
except ImportError:
    install(["python3-requests"])
    import requests


class ActionError(Exception):
    """Common exception for known error conditions."""

    def __init__(self, message):
        """Create an ActionError.

        :param message: Description of the error that has occurred.
        """
        self.message = message


ES_BASE_URL = "http://localhost:9200"


def send_request(path, method="GET", data={}):
    """Fetch <ES_BASE_URL>/<path> using 'method' and return the output.

    :param path: The path to fetch from ES_BASE_URL
    :type path: str

    :param method: The HTTP method to use
    :type method: str

    :param data: Data to send
    :type data: dict

    :raises ActionError

    :return: Returns the text response from the server.
    :rtype: string
    """
    url = urljoin(ES_BASE_URL, path)

    try:
        response = requests.request(method, url, json=data)
        response.raise_for_status()
        return response.text
    except requests.exceptions.ConnectTimeout as e:
        raise ActionError(
            "Request timed out while trying to connect to server: {}".format(str(e))
        )
    except requests.exceptions.ConnectionError as e:
        raise ActionError("Failed to connect to server: {}".format(str(e)))
    except requests.exceptions.ReadTimeout as e:
        raise ActionError(
            "Server did not send any data in the allotted amount of time: {}".format(
                str(e)
            )
        )
    except requests.exceptions.HTTPError as e:
        raise ActionError("An HTTP error occurred: {}".format(str(e)))
    except requests.exceptions.RequestException as e:
        raise ActionError("An error occurred: '{}".format(str(e)))


def get_format_param():
    """Return the action's output format as a query parameter.

    :return: Format query parameter
    :rtype: string
    """
    format = function_get("format")

    if format == "text":
        return "v"
    if format in ["json", "yaml"]:
        return "format=" + format

    function_fail("Unknown format option: {}".format(format))


def httpget_func(path):
    """Compose a function which sends a 'GET' request and sets the action output.

    The format query parameter of the action is added to the path.

    :param path: Path to send the 'GET' request to

    :return: Callable which sends a 'GET' request and sets the action output
    :rtype: Callable
    """

    def set_output_from_httpget(path):
        path = path + "?" + get_format_param()
        output = send_request(path)
        function_set({"output": output})

    return partial(set_output_from_httpget, path)


def list_cluster_settings():
    """Show cluster settings.

    Show default settings if 'include_defaults = True'
    """
    include_defaults = function_get("include-defaults")
    flat_settings = function_get("flat-settings")

    path = "/_cluster/settings?" + get_format_param()
    if flat_settings:
        path += "&flat_settings"
    if include_defaults:
        path += "&include_defaults"

    output = send_request(path)
    function_set({"output": output})


def disable_shard_allocation():
    """Disable shard allocation for all indices."""
    log("Disabling shard allocation")

    path = "/_cluster/settings?" + get_format_param()
    data = {"persistent": {"cluster.routing.allocation.enable": "none"}}
    output = send_request(path, method="PUT", data=data)
    function_set({"output": output})


def enable_shard_allocation():
    """Enable shard allocation for all indices."""
    log("Enabling shard allocation")

    path = "/_cluster/settings?" + get_format_param()
    data = {"persistent": {"cluster.routing.allocation.enable": "all"}}
    output = send_request(path, method="PUT", data=data)
    function_set({"output": output})


def synced_flush():
    """Stop indexing and perfom a syced flush."""
    log("Stopping indexing and perfoming a synced flush")

    path = "/_flush/synced?" + get_format_param()
    output = send_request(path, method="POST")
    function_set({"output": output})


ACTIONS = {
    "get-health": httpget_func("/_cat/health"),
    "list-nodes": httpget_func("/_cat/nodes"),
    "list-indices": httpget_func("/_cat/indices"),
    "list-shards": httpget_func("/_cat/shards"),
    "recovery-status": httpget_func("/_cat/recovery"),
    "version": httpget_func("/"),
    "list-cluster-settings": list_cluster_settings,
    "disable-shard-allocation": disable_shard_allocation,
    "enable-shard-allocation": enable_shard_allocation,
    "synced-flush": synced_flush,
}


def main(args):
    """Execute the appropriate action.

    :param args: Argument list from the invocation.
    :type args: list
    """
    action_name = os.path.basename(args[0])
    try:
        action = ACTIONS[action_name]
    except KeyError:
        s = "Action {} undefined".format(action_name)
        function_fail(s)
        return s
    else:
        try:
            action()
        except ActionError as e:
            s = "Action {} failed with message: {}".format(action_name, e.message)
            function_fail(s)
            return s


if __name__ == "__main__":
    sys.exit(main(sys.argv))
