#!/usr/bin/env python3

"""Script which waits for peers to join an ES cluster."""

import logging
import os
import socket
import sys
import time
import traceback
import functools

logging.basicConfig(
    stream=sys.stderr,
    level=logging.DEBUG,
    format="%(asctime)s [%(levelname)s] %(message)s",
)

import requests  # noqa: E402

sys.path.append(os.path.abspath("mod/charm-helpers"))
from charmhelpers.core.host import service_restart  # noqa: I100,E402
from charmhelpers.core.host import service_running  # noqa: I100,E402


POLL_DELAY = 5
SCRIPT_TIMEOUT = 600
CONNECTION_TIMEOUT = 5


def main():
    """Run the script."""
    start_time = time.time()
    while True:
        node_count = get_node_count()
        if node_count > 1:
            logging.info("Multiple ES nodes detected; success.")
            return 0
        if time.time() - start_time > SCRIPT_TIMEOUT:
            logging.error("Timed out while waiting for peers")
            return 1
        logging.info("Node count is not greater than 1, will restart and retry")
        logging.info("Restarting elasticsearch...")
        restart_elasticsearch()
        logging.info("Waiting for elasticsearch service to start...")
        wait_for_es_service()
        logging.info("Waiting for elasticsearch port to open...")
        wait_for_es_port()
        logging.info("Waiting {} seconds...".format(POLL_DELAY))
        time.sleep(POLL_DELAY)


def get_node_count():
    """Return the number of nodes detected in the cluster.

    ES communication errors are assumed to be temporary, and in such cases this function
    returns 0.

    """
    # NOTE: 503 errors have been intermittently sighted here, so we'll retry
    # in case of errors.
    cluster_health_url = "http://localhost:9200/_cluster/health"
    max_attempts = 5
    for i in range(max_attempts):
        try:
            return requests.get(cluster_health_url, timeout=CONNECTION_TIMEOUT).json()[
                "number_of_nodes"
            ]
        except Exception:
            if i < max_attempts - 1:
                logging.info(
                    "Error occurred while polling {}; retrying".format(
                        cluster_health_url
                    )
                )
                time.sleep(POLL_DELAY)
            else:
                logging.warning(traceback.format_exc())
    else:
        logging.warning(
            "Error occurred while polling {}; giving up".format(cluster_health_url)
        )
        return 0


def restart_elasticsearch():
    """Restart the ElasticSearch service."""
    service_restart("elasticsearch")


def _retry(timeout=60, msg="Waiting for an operation to succeed..."):
    def decoratee(func):
        @functools.wraps(func)
        def _decoratee(*args, **kwargs):
            start_time = time.time()
            while time.time() - start_time < timeout:
                try:
                    func(*args, **kwargs)
                except Exception:
                    time.sleep(POLL_DELAY)
                else:
                    logging.info("{} [Success]".format(msg))
                    return
            # Final trial
            func(*args, **kwargs)

        return _decoratee

    return decoratee


# Note: restarting the service could take more than 60s, let's wait for it
# longer before giving up. Addressing bug #1846366.
@_retry(timeout=180, msg="Waiting for elasticsearch service to start...")
def wait_for_es_service():
    """Wait for the ElasticSearch service to start."""
    if not service_running("elasticsearch"):
        raise Exception("Failed to start elasticsearch service.")


# Note: restarting the service could take more than 60s, let's wait for it
# longer before giving up. Addressing bug #1846366.
@_retry(timeout=180, msg="Waiting for elasticsearch port to open...")
def wait_for_es_port():
    """Wait for the ElasticSearch port to open."""
    _test_socket_connection("localhost", 9200)


def _test_socket_connection(host, port):
    s = socket.create_connection((host, port), timeout=CONNECTION_TIMEOUT)
    s.close()


if __name__ == "__main__":
    sys.exit(main())
